@@ -140,7 +140,7 @@ let sensitiveOps = object
   method mode = Sensitive
   method modeDesc = "case sensitive"
   method compare s s' = compare (s : string) s'
-  method hash s = Hashtbl.hash s
+  method hash s = Uutil.hash s
   method normalizePattern s = s
   method caseInsensitiveMatch = false
   method normalizeMatchedString s = s
@@ -152,7 +152,7 @@ let insensitiveOps = object
   method mode = Insensitive
   method modeDesc = "Latin-1 case insensitive"
   method compare s s' = Util.nocase_cmp s s'
-  method hash s = Hashtbl.hash (String.lowercase s)
+  method hash s = Uutil.hash (String.lowercase s)
   method normalizePattern s = s
   method caseInsensitiveMatch = true
   method normalizeMatchedString s = s
@@ -164,7 +164,7 @@ let unicodeSensitiveOps = object
   method mode = UnicodeSensitive
   method modeDesc = "Unicode case sensitive"
   method compare s s' = Unicode.case_sensitive_compare s s'
-  method hash s = Hashtbl.hash (Unicode.decompose s)
+  method hash s = Uutil.hash (Unicode.decompose s)
   method normalizePattern p = Unicode.decompose p
   method caseInsensitiveMatch = false
   method normalizeMatchedString s = Unicode.decompose s
@@ -176,7 +176,7 @@ let unicodeInsensitiveOps = object
   method mode = UnicodeInsensitive
   method modeDesc = "Unicode case insensitive"
   method compare s s' = Unicode.case_insensitive_compare s s'
-  method hash s = Hashtbl.hash (Unicode.normalize s)
+  method hash s = Uutil.hash (Unicode.normalize s)
   method normalizePattern p = Unicode.normalize p
   method caseInsensitiveMatch = false
   method normalizeMatchedString s = Unicode.normalize s
