@@ -5,7 +5,7 @@ type t
 
 (* Os.safeFingerprint should usually be used rather than these functions *)
 val file : Fspath.t -> Path.local -> t
-val subfile : string -> Int64.t -> Uutil.Filesize.t -> t
+val subfile : Fspath.t -> Int64.t -> Uutil.Filesize.t -> t
 
 val string : string -> t
 
@@ -14,3 +14,6 @@ val toString : t -> string
 (* This dummy fingerprint is guaranteed small and distinct from all
    other fingerprints *)
 val dummy : t
+
+val hash : t -> int
+val equal : t -> t -> bool
