@@ -15,6 +15,11 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)
 
+(*
+See
+http://www.opensource.apple.com/source/copyfile/copyfile-42/copyfile.c
+*)
+
 let debug = Trace.debug "osx"
 
 (****)
@@ -26,7 +31,7 @@ let isMacOSX = isMacOSXPred ()
 (****)
 
 let rsrcSync =
-  Prefs.createString "rsrc" "default"
+  Prefs.createBoolWithDefault "rsrc"
     "!synchronize resource forks (true/false/default)"
     "When set to {\\tt true}, this flag causes Unison to synchronize \
      resource forks and HFS meta-data.  On filesystems that do not \
@@ -45,26 +50,11 @@ let rsrc =
 
 let init b =
   Prefs.set rsrc
-    (Prefs.read rsrcSync = "yes" ||
-     Prefs.read rsrcSync = "true" ||
-     (Prefs.read rsrcSync = "default" && b))
+    (Prefs.read rsrcSync = `True ||
+     (Prefs.read rsrcSync = `Default && b))
 
 (****)
 
-let appleDoubleFile fspath path =
-  let f = Fspath.concatToString fspath path in
-  let len = String.length f in
-  try
-    let i = 1 + String.rindex f '/' in
-    let res = String.create (len + 2) in
-    String.blit f 0 res 0 i;
-    res.[i] <- '.';
-    res.[i + 1] <- '_';
-    String.blit f i res (i + 2) (len - i);
-    res
-  with Not_found ->
-    assert false
-
 let doubleMagic = "\000\005\022\007"
 let doubleVersion = "\000\002\000\000"
 let doubleFiller = String.make 16 '\000'
@@ -126,24 +116,29 @@ let setInt4 v =
   set 0; set 1; set 2; set 3;
   s
 
-let fail path msg =
+let fail dataFspath dataPath doubleFspath msg =
+  debug (fun () -> Util.msg "called 'fail'");
   raise (Util.Transient
-           (Format.sprintf "Malformed AppleDouble file '%s' (%s)" path msg))
+           (Format.sprintf
+              "The AppleDouble Header file '%s' \
+               associated to data file %s is malformed: %s"
+              (Fspath.toPrintString doubleFspath)
+              (Fspath.toPrintString (Fspath.concat dataFspath dataPath)) msg))
 
-let readDouble path inch len =
+let readDouble dataFspath dataPath doubleFspath inch len =
   let buf = String.create len in
   begin try
     really_input inch buf 0 len
   with End_of_file ->
-    fail path "truncated"
+    fail dataFspath dataPath doubleFspath "truncated"
   end;
   buf
 
-let readDoubleFromOffset path inch offset len =
+let readDoubleFromOffset dataFspath dataPath doubleFspath inch offset len =
   LargeFile.seek_in inch offset;
-  readDouble path inch len
+  readDouble dataFspath dataPath doubleFspath inch len
 
-let writeDoubleFromOffset path outch offset str =
+let writeDoubleFromOffset outch offset str =
   LargeFile.seek_out outch offset;
   output_string outch str
 
@@ -154,26 +149,27 @@ let protect f g =
     begin try g () with Sys_error _  | Unix.Unix_error _ -> () end;
     raise e
 
-let openDouble fspath path =
-  let path = appleDoubleFile fspath path in
-  let inch = try open_in_bin path with Sys_error _ -> raise Not_found in
+let openDouble dataFspath dataPath =
+  let doubleFspath = Fspath.appleDouble (Fspath.concat dataFspath dataPath) in
+  let inch =
+    try Fs.open_in_bin doubleFspath with Sys_error _ -> raise Not_found in
   protect (fun () ->
     Util.convertUnixErrorsToTransient "opening AppleDouble file" (fun () ->
-      let header = readDouble path inch 26 in
+      let header = readDouble dataFspath dataPath doubleFspath inch 26 in
       if String.sub header 0 4 <> doubleMagic then
-        fail path "bad magic number";
+        fail dataFspath dataPath doubleFspath "bad magic number";
       if String.sub header 4 4 <> doubleVersion then
-        fail path "bad version";
+        fail dataFspath dataPath doubleFspath "bad version";
       let numEntries = getInt2 header 24 in
       let entries = ref [] in
       for i = 1 to numEntries do
-        let entry = readDouble path inch 12 in
+        let entry = readDouble dataFspath dataPath doubleFspath inch 12 in
         let id = getID entry 0 in
         let ofs = getInt4 entry 4 in
         let len = getInt4 entry 8 in
         entries := (id, (ofs, len)) :: !entries
       done;
-      (path, inch, !entries)))
+      (doubleFspath, inch, !entries)))
     (fun () -> close_in_noerr inch)
 
 (****)
@@ -196,12 +192,13 @@ let ressStampToString r =
         ino mtime ctime (Uutil.Filesize.toString len)
 
 type info =
-  { ressInfo : (string * int64) ressInfo;
+  { ressInfo : (Fspath.t * int64) ressInfo;
     finfo : string }
 
 external getFileInfosInternal :
-      string -> bool -> string * int64 = "getFileInfos"
-external setFileInfosInternal : string -> string -> unit = "setFileInfos"
+  System.fspath -> bool -> string * int64 = "getFileInfos"
+external setFileInfosInternal :
+  System.fspath -> string -> unit = "setFileInfos"
 
 let defaultInfos typ =
   match typ with
@@ -214,7 +211,7 @@ let noTypeCreator = String.make 10 '\000'
 (* Remove trailing zeroes *)
 let trim s =
   let rec trim_rec s pos =
-    if s.[pos - 1] = '\000' then
+    if pos > 0 && s.[pos - 1] = '\000' then
       trim_rec s (pos - 1)
     else
       String.sub s 0 pos
@@ -238,7 +235,7 @@ let extractInfo typ info =
   in
   trim info
 
-let getFileInfos fspath path typ =
+let getFileInfos dataFspath dataPath typ =
   if not (Prefs.read rsrc) then defaultInfos typ else
   match typ with
     (`FILE | `DIRECTORY) as typ ->
@@ -246,7 +243,9 @@ let getFileInfos fspath path typ =
         try
           let (fInfo, rsrcLength) =
             getFileInfosInternal
-              (Fspath.concatToString fspath path) (typ = `FILE) in
+              (Fspath.toSysPath (Fspath.concat dataFspath dataPath))
+              (typ = `FILE)
+          in
           { ressInfo =
               if rsrcLength = 0L then NoRess
               else HfsRess (Uutil.Filesize.ofInt64 rsrcLength);
@@ -254,8 +253,10 @@ let getFileInfos fspath path typ =
         with Unix.Unix_error ((Unix.EOPNOTSUPP | Unix.ENOSYS), _, _) ->
           (* Not a HFS volume.  Look for an AppleDouble file *)
           try
-            let (fspath, path) = Fspath.findWorkingDir fspath path in
-            let (doublePath, inch, entries) = openDouble fspath path in
+            let (workingDir, realPath) =
+              Fspath.findWorkingDir dataFspath dataPath in
+            let (doubleFspath, inch, entries) =
+              openDouble workingDir realPath in
             let (rsrcOffset, rsrcLength) =
               try
                 let (offset, len) = Safelist.assoc `RSRC entries in
@@ -279,22 +280,24 @@ let getFileInfos fspath path typ =
             debug (fun () ->
               Util.msg
                 "AppleDouble for file %s / %s: ressource fork length: %d\n"
-                (Fspath.toString fspath) (Path.toString path)
+                (Fspath.toDebugString dataFspath) (Path.toString dataPath)
                 (Int64.to_int rsrcLength));
             let finfo =
               protect (fun () ->
                 try
                   let (ofs, len) = Safelist.assoc `FINFO entries in
-                  if len < finfoLength then fail doublePath "bad finder info";
-                  readDoubleFromOffset doublePath inch ofs 32
+                  if len < finfoLength then
+                    fail dataFspath dataPath doubleFspath "bad finder info";
+                  readDoubleFromOffset
+                    dataFspath dataPath doubleFspath inch ofs 32
                 with Not_found ->
-                  "")
+                  String.make 32 '\000')
                 (fun () -> close_in_noerr inch)
             in
             close_in inch;
             let stats =
               Util.convertUnixErrorsToTransient "stating AppleDouble file"
-                (fun () -> Unix.LargeFile.stat doublePath) in
+                (fun () -> Fs.stat doubleFspath) in
             { ressInfo =
                 if rsrcLength = 0L then NoRess else
                 AppleDoubleRess
@@ -313,7 +316,7 @@ let getFileInfos fspath path typ =
                    | `Unix  -> 0.
                    end,
                    Uutil.Filesize.ofInt64 rsrcLength,
-                   (doublePath, rsrcOffset));
+                   (doubleFspath, rsrcOffset));
               finfo = extractInfo typ finfo }
           with Not_found ->
             defaultInfos typ)
@@ -334,36 +337,37 @@ let insertInfo fullInfo info =
   String.blit info (offset + 2) fullInfo 24 2;
   fullInfo
 
-let setFileInfos fspath path finfo =
+let setFileInfos dataFspath dataPath finfo =
   assert (finfo <> "");
   Util.convertUnixErrorsToTransient "setting file informations" (fun () ->
     try
-      let (fullFinfo, _) =
-        getFileInfosInternal (Fspath.concatToString fspath path) false in
-      setFileInfosInternal (Fspath.concatToString fspath path)
-        (insertInfo fullFinfo finfo)
+      let p = Fspath.toSysPath (Fspath.concat dataFspath dataPath) in
+      let (fullFinfo, _) = getFileInfosInternal p false in
+      setFileInfosInternal p (insertInfo fullFinfo finfo)
     with Unix.Unix_error ((Unix.EOPNOTSUPP | Unix.ENOSYS), _, _) ->
       (* Not an HFS volume.  Look for an AppleDouble file *)
-      let (fspath, path) = Fspath.findWorkingDir fspath path in
+      let (workingDir, realPath) = Fspath.findWorkingDir dataFspath dataPath in
       begin try
-        let (doublePath, inch, entries) = openDouble fspath path in
+        let (doubleFspath, inch, entries) = openDouble workingDir realPath in
         begin try
           let (ofs, len) = Safelist.assoc `FINFO entries in
-          if len < finfoLength then fail doublePath "bad finder info";
+          if len < finfoLength then
+            fail dataFspath dataPath doubleFspath "bad finder info";
           let fullFinfo =
             protect
               (fun () ->
-                let res = readDoubleFromOffset doublePath inch ofs 32 in
+                let res =
+                  readDoubleFromOffset
+                    dataFspath dataPath doubleFspath inch ofs 32 in
                 close_in inch;
                 res)
               (fun () -> close_in_noerr inch)
           in
           let outch =
-            open_out_gen [Open_wronly; Open_binary] 0o600 doublePath in
+            Fs.open_out_gen [Open_wronly; Open_binary] 0o600 doubleFspath in
           protect
             (fun () ->
-               writeDoubleFromOffset doublePath outch ofs
-                 (insertInfo fullFinfo finfo);
+               writeDoubleFromOffset outch ofs (insertInfo fullFinfo finfo);
                close_out outch)
             (fun () ->
                close_out_noerr outch);
@@ -373,15 +377,17 @@ let setFileInfos fspath path finfo =
                    (Format.sprintf
                       "Unable to set the file type and creator: \n\
                        The AppleDouble file '%s' has no fileinfo entry."
-                      doublePath))
+                      (Fspath.toPrintString doubleFspath)))
         end
       with Not_found ->
         (* No AppleDouble file, create one if needed. *)
         if finfo <> "F" && finfo <> "D" then begin
-          let path = appleDoubleFile fspath path in
+          let doubleFspath =
+            Fspath.appleDouble (Fspath.concat workingDir realPath) in
           let outch =
-            open_out_gen
-              [Open_wronly; Open_creat; Open_excl; Open_binary] 0o600 path
+            Fs.open_out_gen
+              [Open_wronly; Open_creat; Open_excl; Open_binary] 0o600
+              doubleFspath
           in
           (* Apparently, for compatibility with various old versions
              of Mac OS X that did not follow the AppleDouble specification,
@@ -453,7 +459,8 @@ let ressFingerprint fspath path info =
   | AppleDoubleRess (_, _, _, len, (path, offset)) ->
       debug (fun () ->
         Util.msg "ressource fork fingerprint: path %s, offset %d, len %d"
-        path (Int64.to_int offset) (Uutil.Filesize.toInt len));
+        (Fspath.toString path)
+        (Int64.to_int offset) (Uutil.Filesize.toInt len));
       Fingerprint.subfile path offset len
 
 let ressLength ress =
@@ -470,8 +477,8 @@ let openRessIn fspath path =
   Util.convertUnixErrorsToTransient "reading resource fork" (fun () ->
     try
       Unix.in_channel_of_descr
-        (Unix.openfile
-           (Fspath.concatToString fspath (ressPath path))
+        (Fs.openfile
+           (Fspath.concat fspath (ressPath path))
            [Unix.O_RDONLY] 0o444)
     with Unix.Unix_error ((Unix.ENOENT | Unix.ENOTDIR), _, _) ->
       let (doublePath, inch, entries) = openDouble fspath path in
@@ -488,13 +495,13 @@ let openRessOut fspath path length =
   Util.convertUnixErrorsToTransient "writing resource fork" (fun () ->
     try
       Unix.out_channel_of_descr
-        (Unix.openfile
-           (Fspath.concatToString fspath (ressPath path))
+        (Fs.openfile
+           (Fspath.concat fspath (ressPath path))
            [Unix.O_WRONLY;Unix.O_TRUNC] 0o600)
     with Unix.Unix_error ((Unix.ENOENT | Unix.ENOTDIR), _, _) ->
-      let path = appleDoubleFile fspath path in
+      let path = Fspath.appleDouble (Fspath.concat fspath path) in
       let outch =
-        open_out_gen
+        Fs.open_out_gen
           [Open_wronly; Open_creat; Open_excl; Open_binary] 0o600 path
       in
       protect (fun () ->
@@ -507,6 +514,7 @@ let openRessOut fspath path length =
         output_string outch "\000\000\014\176"; (* length *)
         output_string outch "\000\000\000\002"; (* Resource fork *)
         output_string outch "\000\000\014\226"; (* offset *)
+(* FIX: should check for overflow! *)
         output_string outch (setInt4 (Uutil.Filesize.toInt64 length));
                                                 (* length *)
         output_string outch (emptyFinderInfo ());
