@@ -1,5 +1,20 @@
 (* Unison file synchronizer: src/sortri.ml *)
-(* Copyright 1999-2007 (see COPYING for details) *)
+(* Copyright 1999-2009, Benjamin C. Pierce 
+
+    This program is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+*)
+
 
 open Common  
 
@@ -9,7 +24,7 @@ let dbgsort = Util.debug "sort"
 
 let bysize =
   Prefs.createBool "sortbysize" false
-    "list changed files by size, not name"
+    "!list changed files by size, not name"
     ("When this flag is set, the user interface will list changed files "
      ^ "by size (smallest first) rather than by name.  This is useful, for "
      ^ "example, for synchronizing over slow links, since it puts very "
@@ -22,13 +37,13 @@ let bysize =
 
 let newfirst =
   Prefs.createBool "sortnewfirst" false
-    "list new before changed files"
+    "!list new before changed files"
     ("When this flag is set, the user interface will list newly created "
      ^ "files before all others.  This is useful, for example, for checking "
      ^ "that newly created files are not `junk', i.e., ones that should be "
      ^ "ignored or deleted rather than synchronized.")
 
-let sortfirst = Pred.create "sortfirst"
+let sortfirst = Pred.create "sortfirst" ~advanced:true
     ("Each argument to \\texttt{sortfirst} is a pattern \\ARG{pathspec}, "
      ^ "which describes a set of paths.  "
      ^ "Files matching any of these patterns will be listed first in the "
@@ -36,7 +51,7 @@ let sortfirst = Pred.create "sortfirst"
      ^ "The syntax of \\ARG{pathspec} is "
      ^ "described in \\sectionref{pathspec}{Path Specification}.")
 
-let sortlast = Pred.create "sortlast"
+let sortlast = Pred.create "sortlast" ~advanced:true
     ("Similar to \\verb|sortfirst|, except that files matching one of these "
      ^ "patterns will be listed at the very end.")
 
@@ -85,18 +100,18 @@ let sortNewFirst () =
 (* Main sorting functions *)
 
 let shouldSortFirst ri =
-  Pred.test sortfirst (Path.toString ri.path)
+  Pred.test sortfirst (Path.toString ri.path1)
 let shouldSortLast ri =
-  Pred.test sortlast (Path.toString ri.path)
+  Pred.test sortlast (Path.toString ri.path1)
 
-let newItem ri = 
+let newItem ri =
   let newItem1 ri =
     match ri.replicas with
-      Different((_, `Created, _, _), _, _, _) -> true
+      Different diff -> diff.rc1.status = `Created
     | _ -> false in
   let newItem2 ri =
     match ri.replicas with
-      Different(_, (_, `Created, _, _), _, _) -> true
+      Different diff -> diff.rc2.status = `Created
     | _ -> false
   in newItem1 ri || newItem2 ri
 
@@ -115,6 +130,7 @@ let compareReconItems () =
     let cmp = 
       combineCmp [
         pred problematic;
+        pred partiallyProblematic;
         pred shouldSortFirst;
         invertCmp (pred shouldSortLast);
         if newfirst then pred newItem else 0;
@@ -123,10 +139,10 @@ let compareReconItems () =
           let l2 = Common.riLength ri2 in
           if l1<l2 then -1 else if l2<l1 then 1 else 0
          else 0);
-        (compare (Path.toString ri1.path) (Path.toString ri2.path))
+        (compare (Path.toString ri1.path1) (Path.toString ri2.path1))
       ] in
     dbgsort (fun() -> Util.msg "%s <= %s --> %d\n"
-               (Path.toString ri1.path) (Path.toString ri2.path) cmp);
+               (Path.toString ri1.path1) (Path.toString ri2.path1) cmp);
     cmp
 
 let sortReconItems items = Safelist.stable_sort (compareReconItems()) items
