@@ -1,16 +1,31 @@
 (* Unison file synchronizer: src/stasher.ml *)
 (* $I2: Last modified by lescuyer *)
-(* Copyright 1999-2007 (see COPYING for details) *)
+(* Copyright 1999-2009, Benjamin C. Pierce 
 
-(*------------------------------------------------------------------------------------*)
+    This program is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+*)
+
+
+(* --------------------------------------------------------------------------*)
 (* Preferences for backing up and stashing *)
    
 let debug = Util.debug "stasher"
 let verbose = Util.debug "stasher+"
 
 let backuplocation = 
-  Prefs.createString "backuplocation" "central"
-    "where backups are stored ('local' or 'central')"
+  Prefs.createString "backuploc" "central"
+    "!where backups are stored ('local' or 'central')"
     ("This preference determines whether backups should be kept locally, near the "
      ^ "original files, or"
      ^" in a central directory specified by the \\texttt{backupdir} "
@@ -18,8 +33,10 @@ let backuplocation =
      ^"the same directory as the original files, and if set to \\verb|central|,"
      ^" \\texttt{backupdir} will be used instead.")
     
+let _ = Prefs.alias backuplocation "backuplocation"
+
 let backup =
-  Pred.create "backup"
+  Pred.create "backup" ~advanced:true
     ("Including the preference \\texttt{-backup \\ARG{pathspec}} "
      ^ "causes Unison to keep backup files for each path that matches "
      ^ "\\ARG{pathspec}.  These backup files are kept in the "
@@ -33,13 +50,13 @@ let backup =
 let _ = Pred.alias backup "mirror"
 
 let backupnot =
-  Pred.create "backupnot"
+  Pred.create "backupnot" ~advanced:true
     ("The values of this preference specify paths or individual files or"
      ^ " regular expressions that should {\\em not} "
      ^ "be backed up, even if the {\\tt backup} preference selects "
      ^ "them---i.e., "
      ^ "it selectively overrides {\\tt backup}.  The same caveats apply here "
-     ^ "as with {\\tt ignore} and {\tt ignorenot}.")
+     ^ "as with {\\tt ignore} and {\\tt ignorenot}.")
 
 let _ = Pred.alias backupnot "mirrornot"
     
@@ -49,7 +66,7 @@ let shouldBackup p =
     
 let backupprefix =
   Prefs.createString "backupprefix" ".bak.$VERSION."
-    "prefix for the names of backup files"
+    "!prefix for the names of backup files"
     ("When a backup for a file \\verb|NAME| is created, it is stored "
      ^ "in a directory specified by \\texttt{backuplocation}, in a file called "
      ^ "\\texttt{backupprefix}\\verb|NAME|\\texttt{backupsuffix}."
@@ -61,16 +78,22 @@ let backupprefix =
      ^ " This keyword is ignored if it appears in a directory name"
      ^ " in the prefix; if it  does not appear anywhere"
      ^ " in the prefix or the suffix, it will be automatically"
-     ^ " placed at the beginning of the suffix.")
+     ^ " placed at the beginning of the suffix.  "
+     ^ "\n\n"
+     ^ "One thing to be careful of: If the {\\tt backuploc} preference is set "
+     ^ "to {\\tt local}, Unison will automatically ignore {\\em all} files "
+     ^ "whose prefix and suffix match {\\tt backupprefix} and {\\tt backupsuffix}.  "
+     ^ "So be careful to choose values for these preferences that are sufficiently "
+     ^ "different from the names of your real files.")
     
 let backupsuffix =
   Prefs.createString "backupsuffix" ""
-    "a suffix to be added to names of backup files"
+    "!a suffix to be added to names of backup files"
     ("See \\texttt{backupprefix} for full documentation.")
 
 let backups =
   Prefs.createBool "backups" false
-    "keep backup copies of all files (see also 'backup')"
+    "!keep backup copies of all files (see also 'backup')"
     ("Setting this flag to true is equivalent to "
      ^" setting \\texttt{backuplocation} to \\texttt{local}"
      ^" and \\texttt{backup} to \\verb|Name *|.")
@@ -91,7 +114,7 @@ let translateOldPrefs () =
 	
 let maxbackups =
   Prefs.createInt "maxbackups" 2
-    "number of backed up versions of a file"
+    "!number of backed up versions of a file"
     ("This preference specifies the number of backup versions that will "
      ^ "be kept by unison, for each path that matches the predicate "
      ^ "\\verb|backup|.  The default is 2.")
@@ -101,7 +124,7 @@ let _ = Prefs.alias maxbackups "backupversions"
     
 let backupdir =
   Prefs.createString "backupdir" ""
-    "Directory for storing centralized backups"
+    "!directory for storing centralized backups"
     ("If this preference is set, Unison will use it as the name of the "
      ^ "directory used to store backup files specified by "
      ^ "the {\\tt backup} preference, when {\\tt backuplocation} is set"
@@ -110,17 +133,18 @@ let backupdir =
 
 let backupDirectory () =
   Util.convertUnixErrorsToTransient "backupDirectory()" (fun () ->
-    try Fspath.canonize (Some (Unix.getenv "UNISONBACKUPDIR"))
+    try Fspath.canonize (Some (System.getenv "UNISONBACKUPDIR"))
     with Not_found ->
-      try Fspath.canonize (Some (Unix.getenv "UNISONMIRRORDIR"))
+      try Fspath.canonize (Some (System.getenv "UNISONMIRRORDIR"))
       with Not_found ->
 	if Prefs.read backupdir <> ""
 	then Fspath.canonize (Some (Prefs.read backupdir))
-	else Os.fileInUnisonDir "backup")
+	else Fspath.canonize
+               (Some (System.fspathToString (Os.fileInUnisonDir "backup"))))
 
 let backupcurrent =
-  Pred.create "backupcurrent"
-    ("Including the preference \\texttt{-backupcurrent \\ARG{pathspec}} "
+  Pred.create "backupcurr" ~advanced:true
+    ("Including the preference \\texttt{-backupcurr \\ARG{pathspec}} "
      ^" causes Unison to keep a backup of the {\\em current} version of every file "
      ^ "matching \\ARG{pathspec}.  "
      ^" This file will be saved as a backup with version number 000. Such"
@@ -131,8 +155,8 @@ let backupcurrent =
      ^ "\\sectionref{pathspec}{Path Specification}.")
 
 let backupcurrentnot =
-  Pred.create "backupcurrentnot" 
-   "Exceptions to \\verb|backupcurrent|, like the \\verb|ignorenot| preference."
+  Pred.create "backupcurrnot" ~advanced:true
+   "Exceptions to \\verb|backupcurr|, like the \\verb|ignorenot| preference."
 
 let shouldBackupCurrent p =
      (* BCP: removed next line [Apr 2007]: causes ALL mergeable files to be backed
@@ -142,7 +166,10 @@ let shouldBackupCurrent p =
   (let s = Path.toString p in
       Pred.test backupcurrent s && not (Pred.test backupcurrentnot s))
 
-(*------------------------------------------------------------------------------------*)
+let _ = Pred.alias backupcurrent "backupcurrent"
+let _ = Pred.alias backupcurrentnot "backupcurrentnot"
+
+(* ---------------------------------------------------------------------------*)
 
 (* NB: We use Str.regexp here because we need group matching to retrieve
    and increment version numbers from backup file names. We only use
@@ -180,8 +207,8 @@ let backup_rx () =
   else
     raise (Util.Fatal "Either backupprefix or backupsuffix must contain '$VERSION'")
    
-(* We ignore files whose name ends in .unison.bak, since people may still have these lying around
-   from using previous versions of Unison. *)
+(* We ignore files whose name ends in .unison.bak, since people may still have these
+   lying around from using previous versions of Unison. *)
 let oldBackupPrefPathspec = "Name *.unison.bak"
 
 (* This function creates Rx regexps based on the preferences to ignore
@@ -204,10 +231,14 @@ let addBackupFilesToIgnorePref () =
     match dir_rx with 
       None   -> "Regex " ^ full 
     | Some _ -> "Regex " ^ dir in
-  debug (fun () -> 
-     Util.msg "New pattern being added to ignore preferences: %s\n" theRegExp);
+
   Globals.addRegexpToIgnore oldBackupPrefPathspec;
-  Globals.addRegexpToIgnore theRegExp
+  if Prefs.read backuplocation = "local" then begin
+    debug (fun () -> 
+       Util.msg "New pattern being added to ignore preferences (for backup files):\n   %s\n"
+         theRegExp);
+    Globals.addRegexpToIgnore theRegExp
+  end 
 
 (* We use references for functions that compute the prefixes and suffixes
    in order to avoid using functions from the Str module each time we need them. *)
@@ -285,7 +316,8 @@ let backupPath fspath path =
 
   let rec mkdirectories backdir =
     verbose (fun () -> Util.msg
-      "mkdirectories %s %s\n" (Fspath.toString sFspath) (Path.toString backdir));
+      "mkdirectories %s %s\n"
+         (Fspath.toDebugString sFspath) (Path.toString backdir));
     if not (Os.exists sFspath Path.empty) then
       Os.createDir sFspath Path.empty Props.dirDefault;
     match Path.deconstructRev backdir with
@@ -297,7 +329,7 @@ let backupPath fspath path =
 
   let path0 = makeBackupName path 0 in
   let sourceTyp = (Fileinfo.get true fspath path).Fileinfo.typ in
-  let path0Typ = (Fileinfo.get true sFspath path0).Fileinfo.typ in
+  let path0Typ = (Fileinfo.get false sFspath path0).Fileinfo.typ in
 
   if   (   sourceTyp = `FILE && path0Typ = `FILE
        && (Fingerprint.file fspath path) = (Fingerprint.file sFspath path0))
@@ -306,16 +338,16 @@ let backupPath fspath path =
   then begin
     debug (fun()-> Util.msg
       "[%s / %s] = [%s / %s] = %s: no need to back up\n"
-      (Fspath.toString sFspath) (Path.toString path0)
-      (Fspath.toString fspath) (Path.toString path)
+      (Fspath.toDebugString sFspath) (Path.toString path0)
+      (Fspath.toDebugString fspath) (Path.toString path)
       (showContent sourceTyp fspath path));
     None
   end else begin
     debug (fun()-> Util.msg
       "stashed [%s / %s] = %s is not equal to new [%s / %s] = %s (or one is a dir): stash!\n"
-      (Fspath.toString sFspath) (Path.toString path0)
+      (Fspath.toDebugString sFspath) (Path.toString path0)
       (showContent path0Typ sFspath path0)
-      (Fspath.toString fspath) (Path.toString path)
+      (Fspath.toDebugString fspath) (Path.toString path)
       (showContent sourceTyp fspath path));
     let sPath = f 0 in
     (* Make sure the parent directory exists *)
@@ -328,65 +360,69 @@ let backupPath fspath path =
 
 (*------------------------------------------------------------------------------------*)
 	  
-let backup fspath path (finalDisposition : [`AndRemove | `ByCopying]) =
+let backup fspath path (finalDisposition : [`AndRemove | `ByCopying]) arch =
   debug (fun () -> Util.msg
       "backup: %s / %s\n"
-      (Fspath.toString fspath)
+      (Fspath.toDebugString fspath)
       (Path.toString path));
   Util.convertUnixErrorsToTransient "backup" (fun () ->
+    let (workingDir,realPath) = Fspath.findWorkingDir fspath path in
     let disposeIfNeeded() =
       if finalDisposition = `AndRemove then
-        Os.delete fspath path in
-    if not (Os.exists fspath path) then 
+        Os.delete workingDir realPath in
+    if not (Os.exists workingDir realPath) then 
       debug (fun () -> Util.msg
         "File %s in %s does not exist, so no need to back up\n"  
-        (Path.toString path) (Fspath.toString fspath))
+        (Path.toString path) (Fspath.toDebugString fspath))
     else if shouldBackup path then begin
       match backupPath fspath path with
         None -> disposeIfNeeded()
       | Some (backRoot, backPath) ->
           debug (fun () -> Util.msg "Backing up %s / %s to %s in %s\n" 
-              (Fspath.toString fspath) (Path.toString path)
-              (Path.toString backPath) (Fspath.toString backRoot));
+              (Fspath.toDebugString fspath) (Path.toString path)
+              (Path.toString backPath) (Fspath.toDebugString backRoot));
           let byCopying() = 
             let rec copy p backp =
               let info = Fileinfo.get true fspath p in
               match info.Fileinfo.typ with
               | `SYMLINK ->
                   debug (fun () -> Util.msg "  Copying link %s / %s to %s / %s\n"
-                    (Fspath.toString fspath) (Path.toString p)
-                    (Fspath.toString backRoot) (Path.toString backp));
+                    (Fspath.toDebugString fspath) (Path.toString p)
+                    (Fspath.toDebugString backRoot) (Path.toString backp));
                   Os.symlink backRoot backp (Os.readLink fspath p)
               | `FILE ->
                   debug (fun () -> Util.msg "  Copying file %s / %s to %s / %s\n"
-                    (Fspath.toString fspath) (Path.toString p)
-                    (Fspath.toString backRoot) (Path.toString backp));
+                    (Fspath.toDebugString fspath) (Path.toString p)
+                    (Fspath.toDebugString backRoot) (Path.toString backp));
                   Copy.localFile  fspath p  backRoot backp backp 
                     `Copy  info.Fileinfo.desc
                     (Osx.ressLength info.Fileinfo.osX.Osx.ressInfo)  None
               | `DIRECTORY ->
                   debug (fun () -> Util.msg "  Copying directory %s / %s to %s / %s\n"
-                    (Fspath.toString fspath) (Path.toString p)
-                    (Fspath.toString backRoot) (Path.toString backp));
+                    (Fspath.toDebugString fspath) (Path.toString p)
+                    (Fspath.toDebugString backRoot) (Path.toString backp));
                   Os.createDir backRoot backp info.Fileinfo.desc;
                   let ch = Os.childrenOf fspath p in
                   Safelist.iter (fun n -> copy (Path.child p n) (Path.child backp n)) ch
               | `ABSENT -> assert false in
             copy path backPath;
             debug (fun () -> Util.msg "  Finished copying; deleting %s / %s\n"
-              (Fspath.toString fspath) (Path.toString path));
+              (Fspath.toDebugString fspath) (Path.toString path));
             disposeIfNeeded() in
-          try 
-            if finalDisposition = `AndRemove then
-              Os.rename "backup" fspath path backRoot backPath
-            else
+          begin if finalDisposition = `AndRemove then
+            try
+              (*FIX: this does the wrong thing with followed symbolic links!*)
+              Os.rename "backup" workingDir realPath backRoot backPath
+            with Util.Transient _ ->
+              debug (fun () -> Util.msg "Rename failed -- copying instead\n");
               byCopying()
-          with _ -> 
-            debug (fun () -> Util.msg "Rename failed -- copying instead\n");
+          else
             byCopying()
+          end;
+          Update.iterFiles backRoot backPath arch Xferhint.insertEntry
       end else begin
 	debug (fun () -> Util.msg "Path %s / %s does not need to be backed up\n"
-	    (Fspath.toString fspath)
+	    (Fspath.toDebugString fspath)
 	    (Path.toString path));
         disposeIfNeeded()
       end)
@@ -398,7 +434,7 @@ let rec stashCurrentVersion fspath path sourcePathOpt =
     Util.convertUnixErrorsToTransient "stashCurrentVersion" (fun () ->
       let sourcePath = match sourcePathOpt with None -> path | Some p -> p in
       debug (fun () -> Util.msg "stashCurrentVersion of %s (drawn from %s) in %s\n" 
-               (Path.toString path) (Path.toString sourcePath) (Fspath.toString fspath));
+               (Path.toString path) (Path.toString sourcePath) (Fspath.toDebugString fspath));
       let stat = Fileinfo.get true fspath sourcePath in
       match stat.Fileinfo.typ with
 	`ABSENT -> ()
@@ -429,7 +465,10 @@ let rec stashCurrentVersion fspath path sourcePathOpt =
                 (Osx.ressLength stat.Fileinfo.osX.Osx.ressInfo)
                 None
           end)
-      
+
+let _ =
+Update.setStasherFun (fun fspath path -> stashCurrentVersion fspath path None)
+
 (*------------------------------------------------------------------------------------*)    
     
 (* This function tries to find a backup of a recent version of the file at location
@@ -439,7 +478,7 @@ let getRecentVersion fspath path fingerprint =
   debug (fun () ->
     Util.msg "getRecentVersion of %s in %s\n" 
       (Path.toString path) 
-      (Fspath.toString fspath));
+      (Fspath.toDebugString fspath));
   Util.convertUnixErrorsToTransient "getRecentVersion" (fun () ->
     let dir = stashDirectory fspath in
     let rec aux_find i =
@@ -451,13 +490,13 @@ let getRecentVersion fspath path fingerprint =
 	debug (fun () ->
 	  Util.msg "recent version %s found in %s\n" 
 	    (Path.toString path) 
-	    (Fspath.toString dir));
+	    (Fspath.toDebugString dir));
 	Some (Fspath.concat dir path)
       end else
 	if i = Prefs.read maxbackups then begin
 	  debug (fun () ->
 	    Util.msg "No recent version was available for %s on this root.\n"
-	      (Fspath.toString (Fspath.concat fspath path)));
+	      (Fspath.toDebugString (Fspath.concat fspath path)));
 	  None
 	end else
 	  aux_find (i+1)
