@@ -3,23 +3,32 @@
 
 let debug = Util.debug "prefs"
 
-type 'a t = ('a * string list) ref
+type 'a t =
+  { mutable value : 'a; defaultValue : 'a; mutable names : string list;
+    mutable setInProfile : bool }
 
-let read p = fst !p
+let read p = p.value
 
-let set p v = p:=(v, snd !p)
+let set p v = p.setInProfile <- true; p.value <- v
 
-let name p = snd !p
+let overrideDefault p v = if not p.setInProfile then p.value <- v
 
-let rawPref default = ref default
+let name p = p.names
+
+let readDefault p = p.defaultValue
+
+let rawPref default name =
+  { value = default; defaultValue = default; names = [name];
+    setInProfile = false }
 
 (* ------------------------------------------------------------------------- *)
 
 let profileName = ref None
+let profileFiles = ref []
 
 let profilePathname n =
   let f = Util.fileInUnisonDir n in
-  if Sys.file_exists f then f
+  if System.file_exists f then f
   else Util.fileInUnisonDir (n ^ ".prf")
 
 let thePrefsFile () = 
@@ -27,6 +36,18 @@ let thePrefsFile () =
     None -> raise (Util.Transient("No preference file has been specified"))
   | Some(n) -> profilePathname n
 
+let profileUnchanged () =
+  List.for_all
+    (fun (path, info) ->
+       try
+         let newInfo = System.stat path in
+         newInfo.Unix.LargeFile.st_kind = Unix.S_REG &&
+         info.Unix.LargeFile.st_mtime = newInfo.Unix.LargeFile.st_mtime &&
+         info.Unix.LargeFile.st_size = newInfo.Unix.LargeFile.st_size
+       with Unix.Unix_error _ ->
+         false)
+    !profileFiles
+
 (* ------------------------------------------------------------------------- *)
 
 (* When preferences change, we need to dump them out to the file we loaded   *)
@@ -46,8 +67,9 @@ let resetters = ref []
 
 let addresetter f = resetters := f :: !resetters
 
-let resetToDefaults () = Safelist.iter (fun f -> f()) !resetters
-  
+let resetToDefaults () =
+  Safelist.iter (fun f -> f()) !resetters; profileFiles := []
+
 (* ------------------------------------------------------------------------- *)
 
 (* When the server starts up, we need to ship it the current state of all    *)
@@ -56,31 +78,35 @@ let resetToDefaults () = Safelist.iter (fun f -> f()) !resetters
 (* created, a dumper (marshaler) and a loader (parser) are added to the list *)
 (* kept here...                                                              *)
 
-type dumpedPrefs = (string * string) list
+type dumpedPrefs = (string * bool * string) list
 
-let dumpers = ref ([] : (string * (unit->string)) list)
+let dumpers = ref ([] : (string * bool * (unit->string)) list)
 let loaders = ref (Util.StringMap.empty : (string->unit) Util.StringMap.t)
 
-let adddumper name f =
-  dumpers := (name,f) :: !dumpers
+let adddumper name optional f =
+  dumpers := (name,optional,f) :: !dumpers
 
 let addloader name f =
   loaders := Util.StringMap.add name f !loaders
 
-let dump () = Safelist.map (fun (name,f) -> (name, f())) !dumpers
-  
-let load d =
-  begin
-    Safelist.iter
-      (fun (name, dumpedval) ->
-        let loaderfn =
-          try Util.StringMap.find name !loaders
-          with Not_found -> raise (Util.Fatal
-            ("Preference "^name^" not found: inconsistent Unison versions??"))
-        in loaderfn dumpedval)
-      d
-  end
+let dump () = Safelist.map (fun (name, opt, f) -> (name, opt, f())) !dumpers
 
+let load d =
+  Safelist.iter
+    (fun (name, opt, dumpedval) ->
+       match
+         try Some (Util.StringMap.find name !loaders) with Not_found -> None
+       with
+         Some loaderfn ->
+           loaderfn dumpedval
+       | None ->
+           if not opt then
+             raise (Util.Fatal
+                      ("Preference "^name^" not found: \
+                        inconsistent Unison versions??")))
+    d
+
+(* For debugging *)
 let dumpPrefsToStderr() =
   Printf.eprintf "Preferences:\n";
   Safelist.iter
@@ -99,87 +125,173 @@ let dumpPrefsToStderr() =
 (* generate an appropriate usage message.                                    *)
 exception IllegalValue of string
 
+(* aliasMap: prefName -> prefName *)
+let aliasMap = ref (Util.StringMap.empty : string Util.StringMap.t)
+
+let canonicalName nm =
+  try Util.StringMap.find nm !aliasMap with Not_found -> nm
+
+type typ =
+  [`BOOL | `INT | `STRING | `STRING_LIST | `BOOLDEF | `CUSTOM | `UNKNOWN]
+
+(* prefType : prefName -> type *)
+let prefType = ref (Util.StringMap.empty : typ Util.StringMap.t)
+
+let typ nm = try Util.StringMap.find nm !prefType with Not_found -> `UNKNOWN
+
 (* prefs: prefName -> (doc, pspec, fulldoc)                                  *)
 let prefs =
   ref (Util.StringMap.empty : (string * Uarg.spec * string) Util.StringMap.t)
 
+let documentation nm =
+  try
+    let (doc, _, fulldoc) = Util.StringMap.find nm !prefs in
+    if doc <> "" && doc.[0] = '*' then raise Not_found;
+    let basic = doc = "" || doc.[0] <> '!' in
+    let doc =
+      if not basic then
+        String.sub doc 1 (String.length doc - 1)
+      else
+        doc
+    in
+    (doc, fulldoc, basic)
+  with Not_found ->
+    ("", "", false)
+
+let list () =
+  List.sort String.compare
+    (Util.StringMap.fold (fun nm _ l -> nm :: l) !prefType [])
+
 (* aliased pref has *-prefixed doc and empty fulldoc                         *)
 let alias pref newname =
   (* pref must have been registered, so name pref is not empty, and will be *)
   (* found in the map, no need for catching exception                       *)
   let (_,pspec,_) = Util.StringMap.find (Safelist.hd (name pref)) !prefs in
   prefs := Util.StringMap.add newname ("*", pspec, "") !prefs;
-  pref := (fst !pref, newname::(snd !pref))
+  aliasMap := Util.StringMap.add newname (Safelist.hd (name pref)) !aliasMap;
+  pref.names <- newname :: pref.names
 
-let registerPref name pspec doc fulldoc =
+let registerPref name typ pspec doc fulldoc =
   if Util.StringMap.mem name !prefs then
     raise (Util.Fatal ("Preference " ^ name ^ " registered twice"));
-  prefs := Util.StringMap.add name (doc, pspec, fulldoc) !prefs
+  prefs := Util.StringMap.add name (doc, pspec, fulldoc) !prefs;
+  (* Ignore internal preferences *)
+  if doc = "" || doc.[0] <> '*' then
+    prefType := Util.StringMap.add name typ !prefType
 
-let createPrefInternal name default doc fulldoc printer parsefn =
-  let newCell = rawPref (default, [name]) in
-  registerPref name (parsefn newCell) doc fulldoc;
-  adddumper name (fun () -> Marshal.to_string !newCell []);
-  addprinter name (fun () -> printer (fst !newCell));
-  addresetter (fun () -> newCell := (default, [name]));
-  addloader name (fun s -> newCell := Marshal.from_string s 0);
+let createPrefInternal name typ local default doc fulldoc printer parsefn =
+  let newCell = rawPref default name in
+  registerPref name typ (parsefn newCell) doc fulldoc;
+  adddumper name local
+    (fun () -> Marshal.to_string (newCell.value, newCell.names) []);
+  addprinter name (fun () -> printer newCell.value);
+  addresetter
+    (fun () ->
+       newCell.setInProfile <- false; newCell.value <- newCell.defaultValue);
+  addloader name
+    (fun s ->
+       let (value, names) = Marshal.from_string s 0 in
+       newCell.value <- value);
   newCell
 
-let create name default doc fulldoc intern printer =
-  createPrefInternal name default doc fulldoc printer
-    (fun cell -> Uarg.String (fun s -> set cell (intern (fst !cell) s)))
+let create name ?(local=false) default doc fulldoc intern printer =
+  createPrefInternal name `CUSTOM local default doc fulldoc printer
+    (fun cell -> Uarg.String (fun s -> set cell (intern (read cell) s)))
 
-let createBool name default doc fulldoc =
-  createPrefInternal name default doc fulldoc
+let createBool name ?(local=false) default doc fulldoc =
+  let doc = if default then doc ^ " (default true)" else doc in
+  createPrefInternal name `BOOL local default doc fulldoc
     (fun v -> [if v then "true" else "false"])
     (fun cell -> Uarg.Bool (fun b -> set cell b))
 
-let createInt name default doc fulldoc =
-  createPrefInternal name default doc fulldoc
-    (fun v -> [string_of_int v])  
+let createInt name ?(local=false) default doc fulldoc =
+  createPrefInternal name `INT local default doc fulldoc
+    (fun v -> [string_of_int v])
     (fun cell -> Uarg.Int (fun i -> set cell i))
 
-let createString name default doc fulldoc =
-  createPrefInternal name default doc fulldoc 
+let createString name ?(local=false) default doc fulldoc =
+  createPrefInternal name `STRING local default doc fulldoc
     (fun v -> [v])
     (fun cell -> Uarg.String (fun s -> set cell s))
 
-let createStringList name doc fulldoc =
-  createPrefInternal name [] doc fulldoc
+let createFspath name ?(local=false) default doc fulldoc =
+  createPrefInternal name `STRING local default doc fulldoc
+    (fun v -> [System.fspathToString v])
+    (fun cell -> Uarg.String (fun s -> set cell (System.fspathFromString s)))
+
+let createStringList name ?(local=false) doc fulldoc =
+  createPrefInternal name `STRING_LIST local [] doc fulldoc
     (fun v -> v)
-    (fun cell -> Uarg.String (fun s -> set cell (s::(fst !cell))))
+    (fun cell -> Uarg.String (fun s -> set cell (s:: read cell)))
+
+let createBoolWithDefault name ?(local=false) doc fulldoc =
+  createPrefInternal name `BOOLDEF local `Default doc fulldoc
+    (fun v -> [match v with
+                 `True    -> "true"
+               | `False   -> "false"
+               | `Default -> "default"])
+    (fun cell ->
+       Uarg.String
+         (fun s ->
+            let v =
+              match s with
+                "yes" | "true"     -> `True
+              | "default" | "auto" -> `Default
+              | _                  -> `False
+            in
+            set cell v))
 
 (*****************************************************************************)
 (*                      Command-line parsing                                 *)
 (*****************************************************************************)
 
 let prefArg = function
-    Uarg.Bool(_)   -> "   "
-  | Uarg.Int(_)    -> "n  "
+    Uarg.Bool(_)   -> ""
+  | Uarg.Int(_)    -> "n"
   | Uarg.String(_) -> "xxx"
   | _             -> assert false
 
 let argspecs hook =
   Util.StringMap.fold
     (fun name (doc, pspec, _) l ->
-       let desc =
-         if String.length doc > 0 && doc.[0] = '*' then doc else
-         let arg = prefArg pspec in
-         let spaces =
-           String.make (max 2 (18 - String.length (name ^ arg))) ' ' in
-         arg ^ spaces ^ doc
-       in
-       ("-" ^ name, hook name pspec, desc)::l)
+       ("-" ^ name, hook name pspec, "")::l)
     !prefs []
 
-let printUsage usage = Uarg.usage (argspecs (fun _ s -> s)) usage
+let oneLineDocs u =
+  let formatOne name pspec doc p =
+    if not p then "" else
+    let doc = if doc.[0] = '!'
+                then String.sub doc 1 ((String.length doc) - 1)
+                else doc in
+    let arg = prefArg pspec in
+    let arg = if arg = "" then "" else " " ^ arg in
+    let spaces =
+      String.make (max 1 (18 - String.length (name ^ arg))) ' ' in
+    " -" ^ name ^ arg ^ spaces ^ doc ^ "\n" in
+  let formatAll p =
+    String.concat ""
+      (Safelist.rev
+         (Util.StringMap.fold 
+            (fun name (doc, pspec, _) l ->
+               (formatOne name pspec doc
+                  (String.length doc > 0 && doc.[0] <> '*' && p doc)) :: l)
+            !prefs []))
+  in
+    u ^ "\n" 
+  ^ "Basic options: \n"
+  ^ formatAll (fun doc -> doc.[0] <> '!')
+  ^ "\nAdvanced options: \n"
+  ^ formatAll (fun doc -> doc.[0] = '!')
+
+let printUsage usage = Uarg.usage (argspecs (fun _ s -> s))
+                         (oneLineDocs usage)
 
 let processCmdLine usage hook =
   Uarg.current := 0;
   let argspecs = argspecs hook in
   let defaultanonfun _ =
     print_string "Anonymous arguments not allowed\n";
-    Uarg.usage argspecs usage;
+    Uarg.usage argspecs (oneLineDocs usage);
     exit 2
   in
   let anonfun =
@@ -192,9 +304,9 @@ let processCmdLine usage hook =
       Not_found -> defaultanonfun
   in 
   try
-    Uarg.parse argspecs anonfun usage
+    Uarg.parse argspecs anonfun (oneLineDocs usage)
   with IllegalValue str -> 
-    raise(Util.Fatal(Printf.sprintf "%s \n%s\n" usage str))
+    raise(Util.Fatal(Printf.sprintf "%s \n%s\n" (oneLineDocs usage) str))
 
 let parseCmdLine usage =
   processCmdLine usage (fun _ sp -> sp)
@@ -238,13 +350,27 @@ let string2int name string =
    in the same order as in the file. *)
 let rec readAFile filename : (string * int * string * string) list =
   let chan =
-    try open_in (profilePathname filename)
-    with Sys_error _ ->
-      raise(Util.Fatal(Printf.sprintf "Preference file %s not found" filename)) in
+    try
+      let path = profilePathname filename in
+        profileFiles := (path, System.stat path) :: !profileFiles;
+        System.open_in_bin path
+    with Unix.Unix_error _ | Sys_error _ ->
+      raise(Util.Fatal(Printf.sprintf "Preference file %s not found" filename))
+  in
+  let bom = "\xef\xbb\xbf" in (* BOM: UTF-8 byte-order mark *)
   let rec loop lines =
     match (try Some(input_line chan) with End_of_file -> None) with
       None -> close_in chan; parseLines filename lines
-    | Some(theLine) -> loop (theLine::lines) in
+    | Some(theLine) ->
+        let theLine =
+          (* A lot of Windows tools start a UTF-8 encoded file by a
+             byte-order mark.  We skip it. *)
+          if lines = [] && Util.startswith theLine bom then
+            String.sub theLine 3 (String.length theLine - 3)
+          else
+            theLine
+        in
+        loop (theLine::lines) in
   loop []
 
 (* Takes a list of strings in reverse order and yields a list of "parsed lines"
@@ -254,6 +380,7 @@ and parseLines filename lines =
     match lines with
       [] -> res
     | theLine :: rest ->
+        let theLine = Util.removeTrailingCR theLine in
         let l = Util.trimWhitespace theLine in
         if l = "" || l.[0]='#' then
           loop rest (lineNum+1) res
@@ -335,7 +462,7 @@ let printFullDocs () =
 (*****************************************************************************)
 
 let addprefsto = createString "addprefsto" ""
-  "file to add new prefs to"
+  "!file to add new prefs to"
   "By default, new preferences added by Unison (e.g., new \\verb|ignore| \
    clauses) will be appended to whatever preference file Unison was told \
    to load at the beginning of the run.  Setting the preference \
@@ -348,10 +475,12 @@ let addLine l =
       then profilePathname (read addprefsto)
       else thePrefsFile() in
   try
-    debug (fun() -> Util.msg "Adding '%s' to %s\n" l filename);
-    let resultmsg = l ^ "' added to profile " ^ filename in 
+    debug (fun() ->
+      Util.msg "Adding '%s' to %s\n" l (System.fspathToDebugString filename));
+    let resultmsg =
+      l ^ "' added to profile " ^ System.fspathToPrintString filename in
     let ochan =
-      open_out_gen [Open_wronly; Open_append; Open_creat] 0o600 filename
+      System.open_out_gen [Open_wronly; Open_creat; Open_append] 0o600 filename
     in
     output_string ochan l;
     output_string ochan "\n";
@@ -369,4 +498,3 @@ let addLine l =
 let add name value = addLine (name ^ " = " ^ value)
 
 let addComment c = ignore (addLine ("# " ^ c))
-
